// Triangle mesh viewer + corner table + subdivision + smoothing
// Written by Jarek Rossignac June 2006
//import processing.opengl.*;  // uncomment for OpenGL


// ** SETUP **
void setup() { 
  size(600, 600, P3D);        //for OpenGL use: 
//   size(800, 800, OPENGL);  
  PFont font = loadFont("Courier-14.vlw"); textFont(font, 12);  // font for writing labels on screen
  for (int i=0; i<maxnv; i++) {G[i]=new pt(0,0,0); Nv[i]=new vec(0,0,0);};   // init vertices and normals
  for (int i=0; i<maxnt; i++) {Nt[i]=new vec(0,0,0);};                        // init triangler normals
  for (int i=0; i<w; i++) {for (int j=0; j<w; j++) {                          // make a 2D grid of vertices
    G[w*i+j].setTo(width*.8*j/(w-1)+width/10,height*.8*i/(w-1)+height/10,0);}}    
  for (int i=0; i<w-1; i++) {for (int j=0; j<w-1; j++) {                  // define the triangles for the grid
      V[(i*(w-1)+j)*6]=i*w+j;       V[(i*(w-1)+j)*6+2]=(i+1)*w+j;       V[(i*(w-1)+j)*6+1]=(i+1)*w+j+1;
      V[(i*(w-1)+j)*6+3]=i*w+j;     V[(i*(w-1)+j)*6+5]=(i+1)*w+j+1;     V[(i*(w-1)+j)*6+4]=i*w+j+1;    }; };
  computeBox();  C.F.setToPoint(Cbox); C.pullE(); 
  computeO();                                                  // compute the box and the O table
  fn[0]="mesh.vts"; fn[1]="mesh.vts"; fn[2]="mesh.vts"; fn[3]="mesh.vts";
  } 
 
// ** DRAW **
void draw() {
  int r=int(6*Rbox/width);
  background(255); 
  lights(); directionalLight(128,128,128,0,1,0); directionalLight(128,128,128,0,0,1);
  sphereDetail(4); 
  translate(float(width)/2, float(height)/2, 0.0);     // center view wrt window  
  if ((!keyPressed)&&(mousePressed)) {C.pan(); C.pullE(); };
  if ((keyPressed)&&(mousePressed)) {updateView();}; 
  perspective(PI/3.0,width/height,1.0,6.0*Rbox); 
  C1.track(C); C2.track(C1);  C2.apply();                              
   if (showEdges) {stroke(black);} else {noStroke();}; for(int i=0; i<nt; i++) {int col=Mt[i]*255/rings; fill(col,255-col,10); shade(i); };
  if (showVertices) {noStroke(); for (int i=0; i<nv; i++) {
    if (Mv[i]==0) {if (valence[i]==6) {fill(yellow);} else {fill(blue);}; }
     else {int col=255-Mv[i]*255/rings; fill(col,col,col); } G[i].show(r); };};
  if (showLabels) { fill(black); for (int i=0; i<nv; i++) {G[i].label(str(i),labelD); }; };
  if (showNormals) {stroke(blue); showTriNormals();};                // show triangle normals
  noStroke(); fill(100); shade(t(c)); fill(150); showCorner(c);   // show current corner c in blue and paint its triangle green
  stroke(red); showBorder();                                          // show border edges in red  
//    fill(50); noStroke(); C.F.show(r);  stroke(100,255,100); C.showPose();
  }
  
//***      KEY ACTIONS
void keyPressed() {  
  if (key=='a') {flipOrientation=!flipOrientation; print("Will "); if(flipOrientation) {print("not ");}; println("orientation");};  
  if (key=='c') {collapse(c); c=l(c);}; 
  if (key=='f') {flip(c); }; 
  if (key=='g') {fni=(fni+1)%4; println("Will read model "+fn[fni]);};
  if (key=='h') {C.F.setToPoint(Cbox); C.D=Rbox*2; C.U.setTo(0,1,0); C.E.setToPoint(C.F); C.E.addVec(new vec(0,0,1)); C.pullE(); C.pose();};
  if (key=='j') {C.jump(c); C.snapD(); C.pullE();}
  if (key=='l') {println("l("+c+")="+l(c)); if(l(c)!=-1) {c=l(c);}; if (jumps) {C.jump(c);};};
  if (key=='n') {println("n("+c+")="+n(c)); c=n(c);};  
  if (key=='o') {println("o("+c+")="+o(c)); if(o(c)!=-1) {c=o(c);}; if (jumps) {C.jump(c);}; };  
  if (key=='p') {println("p("+c+")="+p(c)); c=p(c);};  
  if (key=='r') {println("r("+c+")="+r(c)); if(r(c)!=-1) {c=r(c);}; if (jumps) {C.jump(c);};}; 
  if (key=='t') {println("t("+c+")="+t(c));};  
  if (key=='u') {C.U.setTo(0,1,0);};
  if (key=='+') {maxr++;};
  if (key=='-') {maxr--; if (maxr<1) {maxr=1;};};
  if (keyCode==LEFT) {left(); right(); left(); if (jumps) {C.jump(c);};};
  if (keyCode==RIGHT) {right(); left(); right(); if (jumps) {C.jump(c);};};
  if (keyCode==DOWN) {back();left(); right(); right(); left();back(); if (jumps) {C.jump(c);}; };
  if (keyCode==UP) {left(); right(); right(); left(); if (jumps) {C.jump(c);};};
  if (key=='A') {saveMesh(); };  
  if (key=='B') {print("Subdividing..."); splitEdges(); bulge(); splitTriangles(); computeO(); computeValence(); computeTriNormals(); println("done");}; // Butterfly 
  if (key=='C') {computeValence(); computeSkeleton(); }; 
  if (key=='D') {computeDistance(maxr); }; 
  if (key=='E') {showEdges=!showEdges; };  
  if (key=='F') {doFlips(); }; 
  if (key=='G') {loadMesh();  computeBox();  computeO(); computeTriNormals(); computeVertexNormals();};  
  if (key=='I') {computeIsolation();}; 
  if (key=='J') {jumps=!jumps; }; 
  if (key=='L') {showLabels=!showLabels; };  
  if (key=='N') {showNormals=!showNormals; };  
  if (key=='O') {computeO();};  
  if (key=='P') {if (showPath) {showPath=false;} else {computePath(); showPath=true;};};  
  if (key=='R') {print("Refining..."); splitEdges(); splitTriangles(); computeO(); computeValence(); computeTriNormals(); println("done");};   // refine mesh in place
  if (key=='S') {computeVertexNormals(); tuck(0.6); computeVertexNormals(); tuck(-0.6);   };  // bi-laplace smoothing
  if (key=='T') {computeVertexNormals(); tuck(0.1); };  
  if (key=='U') {computeVertexNormals(); tuck(-0.1); };  
  if (key=='V') {showVertices=!showVertices; };  
  if (key=='W') {println(nv+" vertices:"); for (int i=0; i<nv; i++) { print("V"+i+" = "); G[i].write();}; println(nt+" triangles:"); for (int i=0; i<nt; i++) { writeTri(i);}; };  
  };
  
void mousePressed() {C.anchor(); C.pose();  };       // record where the cursor was when the mouse was pressed
void mouseReleased() {C.anchor(); C.pose(); };  // reset the view if any key was pressed when mouse was released 

void updateView() {
  if (keyCode==SHIFT) {C.Pan(); C.pullE(); };
  if (key==' ') {C.pose(); C.Pan(); C.pullE(); };
  if (key=='z') {C.pose(); C.zoom(); C.pullE(); };
  if (key=='1') {C.pose(); C.fly(1.0); C.pullE(); };  if (key=='2') {C.pose(); C.fly(-1.0); C.pullE();  };
  if (key=='3') {C.turn(); C.pullE(); };  if (key=='4') {C.pose(); C.Turn(); C.pullE(); };
  if (key=='5') {C.pan(); C.pullE(); }; 
  }

void left() {if(l(c)!=-1) {c=l(c);};};
void right() {if(r(c)!=-1) {c=r(c);};};
void back() {if(o(c)!=-1) {c=o(c);};};

  





